/*****************************************************************
 * display.c
 * Handling of 16x2 LCD display
 * For Rancilio Silvia Espresso machine
 * Target: AVR 8-bit MCU
 * Author: Marcus Nilsson
 * Created: 2010-04-22
 *****************************************************************/

#include <util/delay.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <stdio.h>  // For sprintf

#include "display.h"
#include "time.h"
#include "mem.h"
#include "buttons.h"
#include "level.h"

// PORTD pin definitions
#define RS_PIN      0x04 // 2
#define ENABLE_PIN  0x08 // 3
#define BL_PIN      0x10 // 4
#define RW_PIN      0x20 // 5

#define LCD_CMD_MODE   PORTD &= ~RS_PIN // RS = 0
#define LCD_DATA_MODE  PORTD |= RS_PIN  // RS = 1

#define LCD_LIGHT_ON   PORTD |= BL_PIN  // Backlight on
#define LCD_LIGHT_OFF  PORTD &= ~BL_PIN // Backlight off

extern Time sysTime;

// Local functions

// Send Data or command to display
// mode must be set first
void lcdSend(UINT8 data)
{
	PORTB = data;         // Output data
	PORTD |= ENABLE_PIN;  // ENABLE = 1;
	PORTD &= ~ENABLE_PIN; // ENABLE = 0;
	_delay_ms(0.17);      // Wait for completion
}

// Global functions
void displayInit(void)
{
	LCD_CMD_MODE;
	_delay_ms(10); // Wait for display to initialize
	lcdSend(0x38); // Set small font (5x8 dots), 8-bit mode, 2-line mode 
}

void displayOn(void)
{
	LCD_CMD_MODE;
	lcdSend(0x0C); // Display on, no blink, no cursor
	LCD_LIGHT_ON;  // Turn on backlight
}

void displayOff(void)
{
	LCD_CMD_MODE;
	LCD_LIGHT_OFF; // Turn off backlight
	lcdSend(0x08); // Display off
}

// Line1: 0x00-0x0F
// Line2: 0x40-0x4F
void displaySet(char *line1, char *line2)
{
	UINT8 i;
	char str1[16], str2[16];

	// Copy chars (line1)
	for(i=0; i<16; i++)
	{
		if(line1[i] == '\0')
			break;
		str1[i] = line1[i];	
	}
	// Pad with spaces (line1)
	while(i<16)
	{
		str1[i++] = ' ';
	}

	// Copy chars (line2)
	for(i=0; i<16; i++)
	{
		if(line2[i] == '\0')
			break;
		str2[i] = line2[i];	
	}
	// Pad with spaces (line2)
	while(i<16)
	{
		str2[i++] = ' ';
	}

	// Write line1
	LCD_CMD_MODE;
	lcdSend(0x80); // Set DDRAM address to 0x00 (data=0x80 | address)
	LCD_DATA_MODE;
	for(i=0; i<16; i++)
	{
		lcdSend(str1[i]); // Write 16 chars (line1)
	}
	// Write line2
	LCD_CMD_MODE;
	lcdSend(0xC0); // Set DDRAM addres to 0x40 (data=0x80 | address)
	LCD_DATA_MODE;
	for(i=0; i<16; i++)
	{
		lcdSend(str2[i]); // Write 16 chars (line2)
	}
}

// Show time and autostart time
void displayUpdateStandby(Time *time, StartTime *startTime)
{
	char line1[17];
	char line2[17];
	sprintf(line1, "     %02d:%02d", time->hour, time->minute); // Center text
	if(timeDisabled(startTime))
	{
		line2[0] = '\0';
	}
	else
	{
		sprintf(line2, " Wakeup: %02d:%02d", startTime->hour, startTime->minute);
	}
	displaySet(line1, line2);
}

// Display shows: "Brewing", tempValue, tempSetpoint, elapsedTime, maxTime
// Use one decimal for temp process value and no decimals for setpoint
void displayUpdateBrew(UINT8 elapsedTime, UINT8 autoBrewTime)
{
	char line1[17];
	char line2[17];

	sprintf(line2, "%1.0f%cC (%d)", currTemp, CHAR_DEG, brewTempSP);

	if(brewMode == MODE_MANUAL)
	{
		sprintf(line1, "Brewing %ds", elapsedTime);
	}
	else // Auto
	{
		sprintf(line1, "Brewing %ds (%d)", elapsedTime, autoBrewTime);
	}
	displaySet(line1, line2);
}

// Display shows: "Steam" and temperature value+setpoint for steam
// Use no decimals
void displayUpdateSteam(void)
{
	char line1[17];
	char line2[17];
	strcpy_P(line1, PSTR("Steam"));
	sprintf(line2, "%1.0f%cC (%d)", currTemp, CHAR_DEG, steamTempSP);
	displaySet(line1, line2);
}

// Display shows "Hot water" and temperature value+setpoint
// Use no decimals
void displayUpdateWater(void)
{
	char line1[17];
	char line2[17];
	strcpy_P(line1, PSTR("Hot water"));
	sprintf(line2, "%1.0f%cC (%d)", currTemp, CHAR_DEG, brewTempSP);
	displaySet(line1, line2);
}

// Go to next menu page
void menuNextPage(UINT8 *menuPage, UINT8 menu)
{
	UINT8 pages;
	if(menu == MENU_USER)
		pages = MENU_USER_PAGES;
	else if(menu == MENU_SETTINGS)
		pages = MENU_SETTINGS_PAGES;
	else if(menu == MENU_ADVANCED)
		pages = MENU_ADVANCED_PAGES;
	else
		pages = 0; // Unknown menu
	(*menuPage)++;
	if(*menuPage >= pages)
		*menuPage = 0;	
}

// Print info on display (button == 0)
// Optional: Change value (up/down) in menu (or go to other menu) (button != 0)
// return: TRUE = go to standby mode
BOOL displayUpdateMenu(UINT8 *page, UINT8 *menu, UINT8 button)
{
	// Weekday strings
	const static char weekdays[7][4] = {"Mon","Tue","Wed","Thu","Fri","Sat","Sun"};
	UINT16 countCleanBackup = 0;    // For undo
	UINT8 upHours, upMinutes;
	char displayLine1[17];
	char displayLine2[17];
	switch(*menu)
	{
		case MENU_USER:
		switch(*page)
		{
			case 0:
				// Default display
				// change between auto/manual (kept in ram)
				if(button == BUTTON_UP || button == BUTTON_DOWN)
				{
					if(brewMode == MODE_MANUAL)
						brewMode = MODE_AUTO;
					else
						brewMode = MODE_MANUAL;
				}
				if(button == 0 && waterLow())
				{
					// Water level is low, show warning instead of default display
					strcpy_P(displayLine1, PSTR("Fill water!"));
					displayLine2[0] = '\0';
				}
				else
				{
					sprintf(displayLine1, "%1.1f%cC (%d)", currTemp, CHAR_DEG, brewTempSP);
					if(brewMode == MODE_AUTO)
					{
						strcpy_P(displayLine2, PSTR("Mode: Auto"));
					}
					else
					{
						strcpy_P(displayLine2, PSTR("Mode: Manual"));
					}
				}
			break;
			case 1:
				// Enter standby
				if(button != 0)
				{
					return TRUE; // enter standby
				}
				strcpy_P(displayLine1, PSTR("Enter standby"));
				displayLine2[0] = '\0';
			break;
			case 2:
				// Time since start (hours:minutes)
				// only print (no change)
				getUpTime(&sysStartTime, &upHours, &upMinutes);
				strcpy_P(displayLine1, PSTR("Time since start"));
				sprintf(displayLine2, "%d:%02d", upHours, upMinutes);
			break;
			case 3:
				// Enter settings menu
				if(button != 0)
				{
					*menu = MENU_SETTINGS;
					*page = 0;
				}
				strcpy_P(displayLine1, PSTR("Settings"));
				displayLine2[0] = '\0';
			break;
		}
		break;

		case MENU_SETTINGS:
		switch(*page)
		{
			case 0:
				// Brew temp setpoint (1C steps) (50-99)
				if(button == BUTTON_UP)
				{
					// Increase
					if(brewTempSP < MAX_BREW_TEMP)
						brewTempSP++;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(brewTempSP > MIN_BREW_TEMP)
						brewTempSP--;
				}
				strcpy_P(displayLine1, PSTR("Brew temp"));
				sprintf(displayLine2, "%d%cC", brewTempSP, CHAR_DEG);
			break;
			case 1:
				// Shot time (1s steps)
				if(button == BUTTON_UP)
				{
					// Increase
					if(shotTime < MAX_SHOT_TIME)
						shotTime++;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(shotTime > MIN_SHOT_TIME)
						shotTime--;
				}
				strcpy_P(displayLine1, PSTR("Shot time"));
				sprintf(displayLine2, "%d s", shotTime);
			break;
			case 2:
				// Start time Mon-Fri (30min steps + enable/disable)
				if(button == BUTTON_UP)
				{
					// Increase
					incStartTime(&startMonFri);
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					decStartTime(&startMonFri);
				}
				strcpy_P(displayLine1, PSTR("Start Mon-Fri"));
				if(timeDisabled(&startMonFri))
				{
					strcpy_P(displayLine2, PSTR("DISABLED"));
				}
				else
				{
					sprintf(displayLine2, "%d:%02d", startMonFri.hour, startMonFri.minute);
				}
			break;
			case 3:
				// Start time Sat-Sun (30min steps + enable/disable)
				if(button == BUTTON_UP)
				{
					// Increase
					incStartTime(&startSatSun);
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					decStartTime(&startSatSun);
				}
				strcpy_P(displayLine1, PSTR("Start Sat-Sun"));
				if(timeDisabled(&startSatSun))
				{
					strcpy_P(displayLine2, PSTR("DISABLED"));
				}
				else
				{
					sprintf(displayLine2, "%d:%02d", startSatSun.hour, startSatSun.minute);
				}
			break;
			case 4:
				// Shot counters
				if(button != 0)
				{
					// Reset or undo
					if(countClean == 0)
					{
						countClean = countCleanBackup;
					}
					else
					{
						countCleanBackup = countClean;
						countClean = 0;
					}
				}
				sprintf(displayLine1, "Count: %lu", countTotal);
				sprintf(displayLine2, "Since clean: %d", countClean); // Max 999
			break;
			case 5:
				// Advanced settings
				if(button != 0)
				{
					*menu = MENU_ADVANCED;
					*page = 0;
					saveSettings(); // Save settings in flash
				}
				strcpy_P(displayLine1, PSTR("Advanced"));
				displayLine2[0] = '\0';
			break;
			case 6:
				// Return
				// Go to default display
				if(button != 0)
				{
					*menu = MENU_USER;
					*page = 0;
					saveSettings(); // Save settings in flash
				}
				strcpy_P(displayLine1, PSTR("Return"));
				displayLine2[0] = '\0';
			break;
		}
		break;

		case MENU_ADVANCED:
		switch(*page)
		{
			case 0:
				// Time (Hour)
				if(button == BUTTON_UP)
				{
					// Increase
					if(sysTime.hour < 23)
						sysTime.hour++;
					else
						sysTime.hour = 0;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(sysTime.hour > 0)
						sysTime.hour--;
					else
						sysTime.hour = 23;
				}
				strcpy_P(displayLine1, PSTR("Set hour"));
				sprintf(displayLine2, "%d:%02d %s", sysTime.hour, sysTime.minute, weekdays[sysTime.weekday]);
			break;
			case 1:
				// Time (Minute)
				if(button == BUTTON_UP)
				{
					// Increase
					if(sysTime.minute < 59)
						sysTime.minute++;
					else
						sysTime.minute = 0;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(sysTime.minute > 0)
						sysTime.minute--;
					else
						sysTime.minute = 59;
				}
				strcpy_P(displayLine1, PSTR("Set minute"));
				sprintf(displayLine2, "%d:%02d %s", sysTime.hour, sysTime.minute, weekdays[sysTime.weekday]);
			break;
			case 2:
				// Time (Weekday)
				if(button == BUTTON_UP)
				{
					// Increase
					if(sysTime.weekday < 6)
						sysTime.weekday++;
					else
						sysTime.weekday = 0;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(sysTime.weekday > 0)
						sysTime.weekday--;
					else
						sysTime.weekday = 6;
				}
				strcpy_P(displayLine1, PSTR("Set weekday"));
				sprintf(displayLine2, "%d:%02d %s", sysTime.hour, sysTime.minute, weekdays[sysTime.weekday]);
			break;
			case 3:
				// Steam temp setpoint (1C steps)
				if(button == BUTTON_UP)
				{
					// Increase
					if(steamTempSP < MAX_STEAM_TEMP)
						steamTempSP++;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(steamTempSP > MIN_STEAM_TEMP)
						steamTempSP--;
				}
				strcpy_P(displayLine1, PSTR("Steam temp"));
				sprintf(displayLine2, "%d%cC", steamTempSP, CHAR_DEG);
			break;
			case 4:
				// Standby timeout. (1h steps)
				if(button == BUTTON_UP)
				{
					// Increase
					if(standbyTimeout < MAX_STANDBY_TIMEOUT)
						standbyTimeout++;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(standbyTimeout > MIN_STANDBY_TIMEOUT)
						standbyTimeout--;
				}
				strcpy_P(displayLine1, PSTR("Standby timeout"));
				sprintf(displayLine2, "%d hours", standbyTimeout);
			break;
			case 5:
				// P-value
				if(button == BUTTON_UP)
				{
					// Increase
					if(regP < MAX_P)
						regP++;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(regP > MIN_P)
						regP--;
				}
				strcpy_P(displayLine1, PSTR("Regulator P"));
				sprintf(displayLine2, "%d", regP);
			break;
			case 6:
				// I-value
				if(button == BUTTON_UP)
				{
					// Increase
					if(regI < MAX_I)
						regI += 0.1;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(regI > MIN_I)
						regI -= 0.1;
				}
				strcpy_P(displayLine1, PSTR("Regulator I"));
				sprintf(displayLine2, "%1.1f", regI);
			break;
			case 7:
				// D-value
				if(button == BUTTON_UP)
				{
					// Increase
					if(regD < MAX_D)
						regD += 1;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(regD > MIN_D)
						regD -= 1;
				}
				strcpy_P(displayLine1, PSTR("Regulator D"));
				sprintf(displayLine2, "%1.0f", regD);
			break;
			case 8:
				// Max steam time (1min steps)
				if(button == BUTTON_UP)
				{
					// Increase
					if(maxSteamTime < MAX_STEAM_TIME)
						maxSteamTime++;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(maxSteamTime > MIN_STEAM_TIME)
						maxSteamTime--;
				}
				strcpy_P(displayLine1, PSTR("Max steam time"));
				sprintf(displayLine2, "%d min", maxSteamTime);
			break;
			case 9:
				// Cleaning threshold (10 shot steps)
				if(button == BUTTON_UP)
				{
					// Increase
					if(cleanLimit < MAX_CLEAN_LIMIT)
						cleanLimit += 10;
				}
				else if(button == BUTTON_DOWN)
				{
					// Decrease
					if(cleanLimit > MIN_CLEAN_LIMIT)
						cleanLimit -= 10;
				}
				strcpy_P(displayLine1, PSTR("Cleaning limit"));
				sprintf(displayLine2, "%d shots", cleanLimit);
			break;
			case 10:
				// RS232 log on/off
				if(button != 0)
				{
					logEnabled = logEnabled ? FALSE : TRUE;
				}
				strcpy_P(displayLine1, PSTR("Logging"));
				if(logEnabled)
				{
					strcpy_P(displayLine2, PSTR("ON"));
				}
				else
				{
					strcpy_P(displayLine2, PSTR("OFF"));
				}
			break;
			case 11:
				// Debug information
				sprintf(displayLine1, "%1.2f%cC (%d)", currTemp, CHAR_DEG, brewTempSP);
				sprintf(displayLine2, "Output: %d", debugOutput);
			break;
			case 12:
				// Return
				// Go to default display
				if(button != 0)
				{
					*menu = MENU_USER;
					*page = 0;
					saveAdvancedSettings(); // Save advanced settings in flash
				}
				strcpy_P(displayLine1, PSTR("Return"));
				displayLine2[0] = '\0';
			break;
		}
		break;
	}
	displaySet(displayLine1, displayLine2);
	return FALSE;
}

