 /*****************************************************************
 * heater.c
 * Handling of heating element with PID regulation
 * For Rancilio Silvia Espresso machine
 * Target: AVR 8-bit MCU
 * Author: Marcus Nilsson
 * Created: 2010-04-22
 *****************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>

#include "heater.h"
#include "adc.h"
#include "mem.h"
#include "serial.h"

#define HEATER_PIN 0x01 // Use PORTC[0] for SSR

#define HEATER_OFF  PORTC &= ~HEATER_PIN; // Turn off SSR
#define HEATER_ON   PORTC |= HEATER_PIN;  // Turn on SSR

// Max accumulated error (Iterm)
#define WINDUP_GUARD_GAIN   (100.0)

// The output level where the temp is hold pretty steady
#define REG_OFFSET 11

// Reset iState if moving faster than this in the right direction
#define RESET_I_VELOCITY 0.007

FLOAT32 iState;   // Accumulated error
FLOAT32 lastTemp; // Temp from last reading 

// Turn off heater from timer compare interrupt
ISR(SIG_OUTPUT_COMPARE2)
{
	HEATER_OFF;
}

void initPid(void)
{
	iState = 0;
}

void inline pidUpdate(void)
{
	UINT16 heatTime;
	FLOAT32 setPoint;
	FLOAT32 error, result;
	FLOAT32 windupGaurd;
	FLOAT32 pTerm, iTerm, dTerm;

	// Get current temperature
	currTemp = (FLOAT32)adcValue/(adcCount<<2); // Multiply count by 4
	adcValue = 0;
	adcCount = 0;

	// Get setpoint
	if(heatMode == HEAT_STEAM)
	{
		setPoint = steamTempSP;
	}
	else
	{
		setPoint = brewTempSP;
	}

	// Calc PID (0-1000ms in 4ms steps)
	
	// Determine how badly we are doing
	error = setPoint - currTemp;
	
	// The pTerm is the view from now, the pgain judges 
	// how much we care about error we are this instant.
	pTerm = regP * error;
	
	// iState keeps changing over time; it's 
	// overall "performance" over time, or accumulated error
	iState += error;

	// Modification to prevent overshoots
	// Reset iState if we are moving in the right direction
	// ToDo: Maybe decrease iState when moving in the right direction instead of reset
	// More decrease when moving faster etc.
	if((error > 0 && (currTemp - lastTemp) > RESET_I_VELOCITY) ||
	(error < 0 && (lastTemp - currTemp) > RESET_I_VELOCITY) ||
	(error < 0.02 && error > -0.02))
	{
		iState = 0;
	}
			
	// To prevent the iTerm getting huge despite lots of 
	// error, we use a "windup guard" 
	windupGaurd = WINDUP_GUARD_GAIN / regI;  
	if(iState > windupGaurd)
	{
		iState = windupGaurd;
	} 
	else if(iState < -windupGaurd) 
	{
		iState = (-windupGaurd);
	}
	
	iTerm = regI * iState;
	
	// The dTerm, the difference between the temperature now
	// and our last reading, indicated the "speed" 
	// how quickly the temp is changing. (aka. Differential)
	dTerm = (regD * (currTemp - lastTemp));
	/*
	ToDo: Maybe use different D-gain when heating and cooling
	dTerm = currTemp - lastTemp;
	if(dTerm > 0)
		dTerm *= regD_heating; // Heating up
	else
		dTerm *= regD_cooling; // Cooling down
	*/

	// The feedback part
	result = pTerm + iTerm - dTerm;

	// Modify result with our offset
	// ToDo: We could use a larger offset of 25 or something after a cold start where
	// uptime < 30min and startTemp < 40
	result += REG_OFFSET;

	// Get a value between 0 and 256
	// 0 = OFF for the whole period
	// 256 = ON for the whole period
	// 1 = Heat for 1/256s = 4ms
	// 255 = 255/256s = 996ms
	if(result < 0)
		heatTime = 0;
	else if(result > 256)
		heatTime = 256;
	else
		heatTime = result;

	// Operate heater
	if(heatTime < 1 || heatMode == HEAT_OFF)
	{
		// Off all the time
		HEATER_OFF;
		TIMSK = 0x40; // Disable compare match interrupt
	}
	else if(heatTime > 255)
	{
		// On all the time
		HEATER_ON;
		TIMSK = 0x40; // Disable compare match interrupt
	}
	else
	{
		// On for 4-966ms
		// Set timer2 compare match between 1 and 255
		// and let compare match interrupt turn off heater
		HEATER_ON;
		OCR2 = heatTime; // Set compare value
		TIMSK = 0xC0; // Enable compare match interrupt
	}

	// Save temp for next round
	lastTemp = currTemp;

	// Save output for debug information
	debugOutput = heatTime;

	// Send temperature on serial port
	// Must be fast to not interfere with heating
	if(logEnabled)
	{
		serialSend(currTemp);
	}
}

