/*****************************************************************
 * main.c
 * Main code
 * For Rancilio Silvia Espresso machine
 * Target: AVR 8-bit MCU
 * Author: Marcus Nilsson
 * Created: 2010-04-20
 *
 * MCU: Atmel ATMega16-16PU
 * - DIL-40 package
 * - 8 MHz internal oscillator
 * - 32 IOs
 * - 16 kB flash (program)
 * - 1 kB SRAM (working memory)
 * - 512 B EEPROM (settings)
 * - JTAG interface
 * - 2x8-bit timers
 * - 1x16-bit timer
 * - 10-bit ADC
 * - USART
 * - IRQ
 *
 * External hardware
 * - Temperature sensor National LM35
 * - Power switch
 * - Brew switch
 * - Steam switch
 * - Water switch
 * - Up button
 * - Down button
 * - Enter button
 * - LCD display 2x16 characters
 * - Optical water level sensor
 * - Heater relay (SSR)
 * - Pump relay (2-pole for pump and 3-way valve)
 * - 32kHz Crystal oscillator
 * - Serial RS232 port
 *
 * IO-port configuration
 * A0 - Analog input              - Temp sensor
 * A1 - Digital input with pullup - Enter button
 * A2 - Digital input with pullup - Down button
 * A3 - Digital input with pullup - Up button
 * A4 - Digital input with pullup - Steam switch
 * A5 - Digital input with pullup - Power switch
 * A6 - Digital input with pullup - Brew switch
 * A7 - Digital input with pullup - Water switch
 * B0 - Digital output            - LCD Data[0]
 * B1 - Digital output            - LCD Data[1]
 * B2 - Digital output            - LCD Data[2]
 * B3 - Digital output            - LCD Data[3]
 * B4 - Digital output            - LCD Data[4]
 * B5 - Digital output            - LCD Data[5]
 * B6 - Digital output            - LCD Data[6]
 * B7 - Digital output            - LCD Data[7]
 * C0 - Digital output            - Heater SSR relay
 * C1 - Digital output            - Pump relay
 * C2 - Digital input with pullup - JTAG
 * C3 - Digital input with pullup - JTAG
 * C4 - Digital input with pullup - JTAG
 * C5 - Digital input with pullup - JTAG
 * C6 - Digital input             - TOSC2/Quartz
 * C7 - Digital input             - TOSC1/Quartz
 * D0 - Digital input with pullup - Not connected (Serial RxD)
 * D1 - Digital output            - Serial TxD
 * D2 - Digital output            - LCD RS
 * D3 - Digital output            - LCD Enable
 * D4 - Digital output            - LCD Backlight
 * D5 - Digital output            - LCD R/W
 * D6 - Digital input             - Water level sensor
 * D7 - Digital input with pullup - Not connected
 *
 * Additional pins
 * VCC    - 5V Supply
 * GND    - Ground
 * RESET  - Connected to VCC by 10kOhm resistor and to JTAG
 * AREF   - Connected to ground by 100nF or larger capacitor
 * AVCC   - Not used when using internal reference
 * XTAL1  - Not connected
 * XTAL2  - Not connected
 * 
 * Fuse bits
 * 0=programmed, 1=unprogrammed
 * CKSEL=0100 => 8mhz
 * Set EEAVE bit to zero to keep eeprom when doing chip erase
 * Check JTAGEN, SPIEN, EESAVE, CKSEL3, CKSEL1, CKSEL0
 * Select Int. RC Osc. 8MHz; start-up time: 6 CK + 64ms
 * High fuse byte 0x91, Low fuse byte 0xE4
 *
 * Environment settings
 * Device: atmega16
 * F_CPU=8000000
 * Optimization: -0s
 * Linker options: -Wl,-u,vfprintf
 *
 * Other notes
 * Watchdog is off by default
 * Internal RC oscillator can be calibrated if needed (Now set to 0x9E)
 *
 *****************************************************************/

// AVR Includes
#include <avr/io.h>

// Local includes
#include "def.h"
#include "heater.h"
#include "display.h"
#include "time.h"
#include "mem.h"
#include "pump.h"
#include "buttons.h"
#include "serial.h"
#include "adc.h"

// States
#define STATE_OFF      0
#define STATE_STANDBY  1
#define STATE_IDLE     2
#define STATE_BREW     3
#define STATE_STEAM    4
#define STATE_WATER    5

// Number of seconds to pause after brew stop before going back to idle
// Delay after brew stop
#define BREW_PAUSE_TIME  1

// Minimum brewing time in seconds to increment counters
#define MIN_CUP_TIME 11

// Display light in standby mode (0-255)
#define STANDBY_BRIGHTNESS 70

// Main function
int main(void)
{
	UINT8 menu, menuPage;
	UINT8 buttons;
	BOOL exitState;
	UINT8 state = STATE_OFF;
	Time time;
	Time startTime;
	StartTime *autoStartTime;
	UINT8 autoBrewTime = 0;
	TimeStamp brewStartTime;
	TimeStamp pauseStartTime;
	UINT8 elapsedTime;

	// Machine plugged in

	// Setup IO-ports
	DDRA  = 0x00; // All inputs
	PORTA = 0xFE; // Pullups on PORTA[1-7]
	DDRB  = 0xFF; // All outputs
	PORTB = 0x00; // Output zero volt
	DDRC  = 0x03; // PORTC[0-1] is output, PORTC[2-7] is input
	PORTC = 0x3C; // Pullups on PORTC[2-5]
	DDRD  = 0x3E; // PORTD[1-5] is output, PORTD[0] and PORTD[6-7] is input
	PORTD = 0x81; // Pullups on PORTD[0,7]

	// Setup ADC
	adcSetup();

	// Init display
	displayInit();
	
	// setup serial RS-232 communication
	serialInit();

	// Init memory
	initRam();

	// Init PID regulator
	initPid();

	// Init 1s IRQ
	initTimerIrq();

	while(1)
	{
		exitState = FALSE;
		switch(state)
		{
			case STATE_OFF:
				// Turn off relays
				pumpOff();
				heatMode = HEAT_OFF;
				displayOff(); // Turn off display
				saveCounters();
				while((getButtons() & BUTTON_POWER) == 0)
				{
					// Timekeeping. Do nothing.
				}
				// Power on
				getTime(&sysStartTime); // Save start time
				initPid();
				displaySet("",""); // Clear display
				displayOn();
				
				if((getButtons() & BUTTON_MENU_BUTTONS) == (BUTTON_UP | BUTTON_DOWN))
				{
					// Make a factory reset by holding up and down buttons when
					// switching on power button
					displaySet("Factory reset?", "");
					secDelay(6);
					if((getButtons() & BUTTON_MENU_BUTTONS) == (BUTTON_UP | BUTTON_DOWN)) // Still holding buttons?
					{
						factoryReset();
						displaySet("Reset done!", "");
					}
					else
					{
						displaySet("Aborted!", "");
					}
					secDelay(3);
				}

				state = STATE_IDLE;
				break;
			case STATE_STANDBY:
				// Turn off relays
				pumpOff();
				heatMode = HEAT_OFF;
				// Dim down dsplay
				// Start PWM on LCD backlight pin OC1B using timer 1
				TCCR1A=0x21; // 8-bit operation
				OCR1B = STANDBY_BRIGHTNESS;
				TCCR1B=0x0B; // Prescaler 64 -> 488Hz
				while(1)
				{
					getTime(&time);
					getStartTime(&autoStartTime, &time);
					displayUpdateStandby(&time, autoStartTime); // Show time and autostart time
					buttons = getButtons();
					if((buttons & BUTTON_POWER) == 0)
					{
						state = STATE_OFF;
						break;
					}
					else if(buttons & BUTTON_ENTER || startNow(&time, autoStartTime))
					{
						sysStartTime = time; // Save start time
						initPid();
						state = STATE_IDLE;
						break;
					}
				}
				// Stop PWM on LCD backlight and go back to full light
				TCCR1A = 0x00;
				TCCR1B = 0x00;
				break;
			case STATE_IDLE:
				pumpOff();
				heatMode = HEAT_NORMAL;
				waitForMenuButtonsRelease();
				if(getButtons() & BUTTON_ANY_BUT_POWER)
				{
					displaySet("Release buttons", "");
					while(getButtons() & BUTTON_ANY_BUT_POWER)
					{
						// Wait..
						if((getButtons() & BUTTON_POWER) == 0)
						{
							state = STATE_OFF;
							exitState = TRUE;
							break;
						}
					}
				}
				if(exitState)
				{
					break;
				}
				if(countClean >= cleanLimit)
				{
					displaySet("Clean machine!", ""); // might include countClean!
					secDelay(5);
				}
				getTime(&startTime); // Save idle start time
				menu = MENU_USER;
				menuPage = 0;
				while(1) // While in idle mode
				{
					displayUpdateMenu(&menuPage, &menu, 0);
					buttons = getButtons();
					if((buttons & BUTTON_POWER) == 0)
					{
						state = STATE_OFF;
						break;
					}
					else if(buttons & BUTTON_ENTER)
					{
						menuNextPage(&menuPage, menu);
						displayUpdateMenu(&menuPage, &menu, 0);
						getTime(&startTime);
						waitForMenuButtonsRelease();
					}
					else if(buttons & BUTTON_UP)
					{
						if(displayUpdateMenu(&menuPage, &menu, BUTTON_UP))
						{
							state = STATE_STANDBY;
							break;
						}
						getTime(&startTime);
						waitForMenuButtonsRelease();
					}
					else if(buttons & BUTTON_DOWN)
					{
						if(displayUpdateMenu(&menuPage, &menu, BUTTON_DOWN))
						{
							state = STATE_STANDBY;
							break;
						}
						getTime(&startTime);
						waitForMenuButtonsRelease();
					}
					else if(buttons & BUTTON_BREW)
					{
						state = STATE_BREW;
						break;
					}
					else if(buttons & BUTTON_STEAM)
					{
						state = STATE_STEAM;
						break;
					}
					else if(buttons & BUTTON_WATER)
					{
						state = STATE_WATER;
						break;
					}
					// Go to standby after timeout (configurable time)
					// when no buttons were pressed within timeout
					if(idleTimeout(startTime))
					{
						state = STATE_STANDBY;
						break;
					}
				}
				break;	
			case STATE_BREW:
				heatMode = HEAT_NORMAL;
				autoBrewTime = shotTime;
				pumpOn();
				getTimeStamp(&brewStartTime);
				exitState = FALSE;
				while(1)
				{
					elapsedTime = getElapsedSeconds(&brewStartTime);
					displayUpdateBrew(elapsedTime, autoBrewTime); // "Brewing", tempValue, tempSetpoint, elapsedTime, maxTime (if not manual)
					if(brewMode != MODE_MANUAL && elapsedTime >= autoBrewTime)
					{
						// Timeout
						pumpOff();
						incCounters();
						while(getButtons() & BUTTON_BREW)
						{
							// Wait for button release
							// Check power switch also
							if((getButtons() & BUTTON_POWER) == 0)
							{
								exitState = TRUE; // Do not go to idle
								state = STATE_OFF;
								break;
							}
						}
						if(!exitState)
						{
							state = STATE_IDLE; // Go back to idle after brewbutton release
						}
						break;
					}
					buttons = getButtons();
					// Check power switch also
					if((buttons & BUTTON_POWER) == 0)
					{
						state = STATE_OFF;
						break;
					}
					else if((buttons & BUTTON_BREW) == 0)
					{
						// Paused
						pumpOff();
						if(elapsedTime >= MIN_CUP_TIME)
						{
							incCounters();
						}
						getTimeStamp(&pauseStartTime);
						while(1)
						{
							elapsedTime = getElapsedSeconds(&pauseStartTime);
							if(elapsedTime >= BREW_PAUSE_TIME)
							{
								exitState = TRUE;
								state = STATE_IDLE;
								break; // go to idle after pause
							}
							buttons = getButtons();
							// Check power switch also
							if((buttons & BUTTON_POWER) == 0)
							{
								exitState = TRUE;
								state = STATE_OFF;
								break;
							}
							else if(buttons & BUTTON_BREW)
							{
								// Start over again
								pumpOn();
								getTimeStamp(&brewStartTime);
								break; // continue outer loop
							}
						}
						// continue or break
						if(exitState)
						{
							break;
						}
					}
				}
				break;
			case STATE_STEAM:
				pumpOff();
				heatMode = HEAT_STEAM;
				getTime(&startTime);
				while(1)
				{
					displayUpdateSteam(); // Display shows: "Steam" and temperature value+setpoint
					buttons = getButtons();
					if((buttons & BUTTON_POWER) == 0)
					{
						state = STATE_OFF;
						break;
					}
					else if((buttons & BUTTON_STEAM) == 0)
					{
						state = STATE_IDLE;
						break;
					}
					if(steamTimeout(startTime))
					{
						// Timeout
						displaySet("Steam timeout", "");
						heatMode = HEAT_NORMAL;
						// Wait for button release
						while(1)
						{
							buttons = getButtons();
							if((buttons & BUTTON_POWER) == 0)
							{
								state = STATE_OFF;
								break;
							}
							else if((buttons & BUTTON_STEAM) == 0)
							{
								state = STATE_IDLE;
								break;
							}
						}
						break;
					}
				}
				break;
			case STATE_WATER:
				pumpOn();
				heatMode = HEAT_NORMAL;
				while(1)
				{
					displayUpdateWater(); // Display shows "Hot water" and temperature value+setpoint
					buttons = getButtons();
					if((buttons & BUTTON_POWER) == 0)
					{
						state = STATE_OFF;
						break;
					}
					else if((buttons & BUTTON_WATER) == 0)
					{
						state = STATE_IDLE;
						break;
					}
				}
				break;
		}
	}
}


