/*****************************************************************
 * mem.c
 * Memory handling
 * For Rancilio Silvia Espresso machine
 * Target: AVR 8-bit MCU
 * Author: Marcus Nilsson
 * Created: 2011-01-25
 *****************************************************************/

#include <avr/eeprom.h>

#include "mem.h"
#include "heater.h"
#include "time.h"

// Globals

// Only in RAM
UINT8 heatMode;            // off/normal/steam
UINT8 brewMode;            // Auto/manual
Time sysStartTime;         // System start time
FLOAT32 currTemp;          // Current temperature
BOOL logEnabled;           // Enable serial logging
UINT32 adcValue;
UINT16 adcCount;
UINT16 debugOutput;

// Also in EEPROM
// Settings
UINT8 brewTempSP;          // Setpoint for coffee
UINT8 shotTime;            // Shot time
StartTime startMonFri;     // Automatic start for weekdays
StartTime startSatSun;     // Automatic start for weekends
// Advanced settings
UINT8 steamTempSP;         // Setpoint for steam
UINT8 standbyTimeout;      // Standby timeout
UINT8 regP;                // Regulator P-gain parameter
FLOAT32 regI;              // Regulator I-gain parameter
FLOAT32 regD;              // Regulator D-gain parameter
UINT8 maxSteamTime;        // Maximum allowed steam time
UINT16 cleanLimit;         // Cleaning interval
// Counters
UINT32 countTotal;         // Total number of shots
UINT16 countClean;         // Number of shots since clean

void initRam(void)
{
	// Init RAM only values
	heatMode = HEAT_OFF;
	brewMode = MODE_MANUAL;
	logEnabled = FALSE; // Disable serial temp logging
	timeReset();

	// Read all settings from EEPROM to RAM (only at this place, otherwise RAM values are used)

	// Read counters
	eeprom_read_block(&countTotal, EE_COUNT_TOTAL, 4);
	eeprom_read_block(&countClean, EE_COUNT_CLEAN, 2);

	// Read settings
	eeprom_read_block(&brewTempSP, EE_BREW_TEMP, 1);
	eeprom_read_block(&shotTime, EE_SHOT_TIME, 1);
	eeprom_read_block(&startMonFri, EE_START_MONFRI, 2);
	eeprom_read_block(&startSatSun, EE_START_SATSUN, 2);

	// Read advanced settings
	eeprom_read_block(&steamTempSP, EE_STEAM_TEMP, 1);
	eeprom_read_block(&standbyTimeout, EE_STANDBY_TIMEOUT, 1);
	eeprom_read_block(&regP, EE_P, 1);
	eeprom_read_block(&regI, EE_I, 4);
	eeprom_read_block(&regD, EE_D, 4);
	eeprom_read_block(&maxSteamTime, EE_STEAM_TIME, 1);
	eeprom_read_block(&cleanLimit, EE_CLEAN_LIMIT, 2);
}

// Save settings in EEPROM
void saveSettings(void)
{
	eeprom_update_block(&brewTempSP, EE_BREW_TEMP, 1);
	eeprom_update_block(&shotTime, EE_SHOT_TIME, 1);
	eeprom_update_block((UINT8*) &startMonFri, EE_START_MONFRI, 2);
	eeprom_update_block((UINT8*) &startSatSun, EE_START_SATSUN, 2);
}

// Save advanced settings in EEPROM
void saveAdvancedSettings(void)
{
	eeprom_update_block(&steamTempSP, EE_STEAM_TEMP, 1); // Setpoint for steam
	eeprom_update_block(&standbyTimeout, EE_STANDBY_TIMEOUT, 1);
	eeprom_update_block(&regP, EE_P, 1);
	eeprom_update_block(&regI, EE_I, 4);
	eeprom_update_block(&regD, EE_D, 4);
	eeprom_update_block(&maxSteamTime, EE_STEAM_TIME, 1);
	eeprom_update_block(&cleanLimit, EE_CLEAN_LIMIT, 2);
}

// Save counters in EEPROM
void saveCounters(void)
{
	eeprom_update_block(&countTotal, EE_COUNT_TOTAL, 4);
	eeprom_update_block(&countClean, EE_COUNT_CLEAN, 2);
}

// Icrease shot counters
void incCounters(void)
{
	countTotal++;
	countClean++;
}

// Set default values
void factoryReset(void)
{
	brewTempSP = 97;
	shotTime = 22;
	startMonFri.hour = 0xFF;
	startMonFri.minute = 0;
	startSatSun.hour = 0xFF;
	startSatSun.minute = 0;
	steamTempSP = 135;
	standbyTimeout = 3;
	regP = 12;
	regI = 0.4;
	regD = 120;
	maxSteamTime = 5;
	cleanLimit = 100;
	countTotal = 0;
	countClean = 0;

	saveSettings();
	saveAdvancedSettings();
	saveCounters();
}
