/*****************************************************************
 * time.c
 * Time handling
 * For Rancilio Silvia Espresso machine
 * Target: AVR 8-bit MCU
 * Author: Marcus Nilsson
 * Created: 2010-04-22
 *****************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>

#include "time.h"
#include "heater.h"
#include "mem.h"

Time sysTime;   // Clock

// 1s timer interrupt (Timer2 driven by quartz)
ISR(SIG_OVERFLOW2)
{
	pidUpdate();  // Update PID regulator
	timeUpdate(); // Update time
}

// Init interrupts
void initTimerIrq(void)
{
	// Activate 1s IRQ
	// Set up timer, enable timer overflow interrupt
	// Clock source is 32767Hz crystal
	// Prescaler 128
	// Overflow at 256 gives 1s ovf interrupts

	// Use timer2 (8-bit)
	ASSR  = 0x08; // Use crystal
	TCCR2 = 0x05; // Prescaler 128
	TIMSK = 0x40; // Use overflow interrupt
	sei();        // Enable interrupts
}

void timeReset(void)
{
	sysTime.hour = 0;
	sysTime.minute = 0;
	sysTime.second = 0;
	sysTime.weekday = 0;
}

// Once per second
// Hour(24), minute(60), seconds(60), weekday(7)
void inline timeUpdate(void)
{
	if(sysTime.second < 59)
	{
		sysTime.second++;
	}
	else
	{
		sysTime.second = 0;
		if(sysTime.minute < 59)
		{
			sysTime.minute++;
		}
		else
		{
			sysTime.minute = 0;
			if(sysTime.hour < 23)
			{
				sysTime.hour++;
			}
			else
			{
				sysTime.hour = 0;
				if(sysTime.weekday < 6)
				{
					sysTime.weekday++;
				}
				else
				{
					sysTime.weekday = 0;
				}
			}
		}
	}
}

void getTime(Time *time)
{
	UINT8 s = sysTime.second;
	*time = sysTime;
	if(s != sysTime.second)
	{
		// There was an interrupt. Make a safe reading.
		*time = sysTime;
	}
}

// Get current value
void getTimeStamp(TimeStamp *t)
{
	t->second = sysTime.second;
	t->counter = TCNT2;
	if(t->second != sysTime.second)
	{
		// There was an interrupt. Make a safe reading.
		t->second = sysTime.second;
		t->counter = TCNT2;
	}
}

// 0-59 seconds
UINT8 getElapsedSeconds(TimeStamp *fromTime)
{
	TimeStamp now;
	UINT8 elapsedSeconds;
	getTimeStamp(&now);
	if(now.second >= fromTime->second)
	{
		elapsedSeconds = now.second - fromTime->second;
	}
	else
	{
		elapsedSeconds = 60 - fromTime->second + now.second;
	}

	if(now.counter < fromTime->counter)
	{
		elapsedSeconds--; // Not there yet
	}
	return elapsedSeconds;
}

// Time to start?
// Handles also disabled autoStartTime
BOOL startNow(Time *t, StartTime *st)
{
	return t->hour == st->hour && t->minute == st->minute;
}

// Max minutes passed? (maxSteamTime)
// Will be executed several times per second
// Compare minute and second
BOOL steamTimeout(Time startTime)
{
	Time now;
	getTime(&now);
	startTime.minute += maxSteamTime;
	if(startTime.minute > 59)
	{
		startTime.minute -= 60;
	}
	return now.minute == startTime.minute && now.second == startTime.second;
}

// Max hours passed? standbyTimeout
// Compare hours and minutes
BOOL idleTimeout(Time startTime)
{
	Time now;
	getTime(&now);
	startTime.hour += standbyTimeout;
	if(startTime.hour > 23)
	{
		startTime.hour -= 24;
	}
	return now.hour == startTime.hour && now.minute == startTime.minute;
}

// Wait some seconds
// 1-8s
void secDelay(UINT8 sec)
{
	// Setup timer 1
	TCNT1 = 0x0000;            // reset
	TCCR1B = 0x05;             // Internal RC osc with prescaler 1024 (start timer)
	OCR1A = 7813*sec;          // set compare value = 7813*sec
	TIFR |= 0x10;              // clear compare match flag
	while((TIFR & 0x10) == 0); // wait for compare match flag
	TCCR1B = 0x00;             // no clock (stop timer)
}

// Get time since start
void getUpTime(Time *startTime, UINT8 *hours, UINT8 *minutes)
{
	INT8 d, h, m;
	Time now;
	getTime(&now);

	d = now.weekday - startTime->weekday;
	if(d < 0)
		d += 7;

	h = now.hour - startTime->hour;
	if(h < 0)
	{
		h += 24;
		d--;
	}

	m = now.minute - startTime->minute;
	if(m < 0)
	{
		m += 60;
		h--;
	}

	h += (d*24);

	*hours = h;
	*minutes = m;
}

// Start time disabled?
BOOL timeDisabled(StartTime *t)
{
	return t->hour == 0xFF;
}

// Increment start time
// 30 min steps and disabling
void incStartTime(StartTime *t)
{
	if(t->hour == 23 && t->minute == 30)
	{
		t->hour = 0xFF; // Disable
	}
	else if(t->hour == 0xFF)
	{
		t->hour = 0;
		t->minute = 0;
	}
	else
	{
		if(t->minute == 0)
		{
			t->minute = 30;
		}
		else
		{
			t->hour++;
			t->minute = 0;
		}
	}
}

// Decrement start time
// 30 min steps and disabling
void decStartTime(StartTime *t)
{
	if(t->hour == 0 && t->minute == 0)
	{
		t->hour = 0xFF; // Disable
	}
	else if(t->hour == 0xFF)
	{
		t->hour = 23;
		t->minute = 30;
	}
	else
	{
		if(t->minute == 30)
		{
			t->minute = 0;
		}
		else
		{
			t->hour--;
			t->minute = 30;
		}
	}
}

// Get next auto start time
void getStartTime(StartTime **start, Time *t)
{
	if((t->weekday == 6 && t->hour < startSatSun.hour && t->minute < startSatSun.minute)
	|| (t->weekday == 4 && t->hour > startMonFri.hour && t->minute > startMonFri.minute)
	|| t->weekday == 5)
	{
		*start = &startSatSun;
	}
	else
	{
		*start = &startMonFri;
	}
}
